from abc import abstractmethod
from marshmallow import Schema, fields


class BaseSchema(Schema):
    schema_namespace = fields.Str(allow_none=True)

    @abstractmethod
    def make_obj(self, data, **kwargs):
        ...



from marshmallow import fields, post_load
from skatepark.schemas import BaseSchema, PipelineRunSchema


class PipelineMemorySchema(BaseSchema):
    class Meta:
        ordered = True

    type = fields.Str(required=True)
    runs = fields.List(fields.Nested(PipelineRunSchema()))

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.memory import PipelineMemory

        return PipelineMemory(**data)



from skatepark.drivers.prompt.prompt_driver import PromptDriver
from skatepark.drivers.prompt.openai_prompt_driver import OpenAiPromptDriver
from skatepark.drivers.prompt.cohere_prompt_driver import CoherePromptDriver
from skatepark.drivers.memory.memory_driver import MemoryDriver
from skatepark.drivers.memory.disk_memory_driver import DiskMemoryDriver

__all__ = [
    "PromptDriver",
    "OpenAiPromptDriver",
    "CoherePromptDriver",

    "MemoryDriver",
    "DiskMemoryDriver"
]



import os
from typing import Optional
from attr import define, field, Factory
from jinja2 import Environment, FileSystemLoader
from skatepark.utils import TiktokenTokenizer, Tokenizer
import skatepark


@define(frozen=True)
class J2:
    template_name: Optional[str] = field(default=None)
    templates_dir: str = field(default=os.path.join(skatepark.PACKAGE_ABS_PATH, "templates"), kw_only=True)
    tokenizer: Tokenizer = field(default=TiktokenTokenizer(), kw_only=True)
    environment: Environment = field(
        default=Factory(
            lambda self: Environment(
                loader=FileSystemLoader(self.templates_dir),
                trim_blocks=True,
                lstrip_blocks=True
            ),
            takes_self=True
        ),
        kw_only=True
    )

    def render(self, **kwargs):
        if not kwargs.get("stop_sequence"):
            kwargs["stop_sequence"] = self.tokenizer.stop_sequence

        return self.environment.get_template(self.template_name).render(kwargs)

    def render_from_string(self, value: str, **kwargs):
        if not kwargs.get("stop_sequence"):
            kwargs["stop_sequence"] = self.tokenizer.stop_sequence

        return self.environment.from_string(value).render(kwargs)



from __future__ import annotations
import json
from typing import TYPE_CHECKING, Optional
from attr import define, field, Factory
from skatepark.memory import PipelineRun
from skatepark.utils import J2

if TYPE_CHECKING:
    from skatepark.drivers import MemoryDriver
    from skatepark.structures import Pipeline


@define
class PipelineMemory:
    type: str = field(default=Factory(lambda self: self.__class__.__name__, takes_self=True), kw_only=True)
    driver: Optional[MemoryDriver] = field(default=None, kw_only=True)
    runs: list[PipelineRun] = field(factory=list, kw_only=True)
    pipeline: Pipeline = field(init=False)

    def add_run(self, run: PipelineRun) -> PipelineMemory:
        self.before_add_run()
        self.process_add_run(run)
        self.after_add_run()

        return self

    def before_add_run(self) -> None:
        pass

    def process_add_run(self, run: PipelineRun) -> None:
        self.runs.append(run)

    def after_add_run(self) -> None:
        if self.driver:
            self.driver.store(self)

    def is_empty(self) -> bool:
        return not self.runs

    def to_prompt_string(self, last_n: Optional[int] = None) -> str:
        return J2("prompts/memory.j2").render(
            runs=self.runs if last_n is None else self.runs[-last_n:]
        )

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), indent=2)

    def to_dict(self) -> dict:
        from skatepark.schemas import PipelineMemorySchema

        return PipelineMemorySchema().dump(self)

    @classmethod
    def from_dict(cls, memory_dict: dict) -> PipelineMemory:
        from skatepark.schemas import PipelineMemorySchema

        return PipelineMemorySchema().load(memory_dict)

    @classmethod
    def from_json(cls, memory_json: str) -> PipelineMemory:
        return PipelineMemory.from_dict(json.loads(memory_json))



from skatepark.schemas.base_schema import BaseSchema

from skatepark.schemas.polymorphic_schema import PolymorphicSchema

from skatepark.schemas.rule_schema import RuleSchema

from skatepark.schemas.tokenizers.tiktoken_tokenizer_schema import TiktokenTokenizerSchema

from skatepark.schemas.drivers.prompt_driver_schema import PromptDriverSchema
from skatepark.schemas.drivers.openai_prompt_driver_schema import OpenAiPromptDriverSchema

from skatepark.schemas.steps.step_schema import StepSchema
from skatepark.schemas.steps.prompt_step_schema import PromptStepSchema
from skatepark.schemas.steps.toolkit_step_schema import ToolkitStepSchema

from skatepark.schemas.summarizers.summarizer_schema import SummarizerSchema
from skatepark.schemas.summarizers.prompt_driver_summarizer_schema import PromptDriverSummarizerSchema

from skatepark.schemas.memory.pipeline_run_schema import PipelineRunSchema
from skatepark.schemas.memory.pipeline_memory_schema import PipelineMemorySchema
from skatepark.schemas.memory.buffer_pipeline_memory_schema import BufferPipelineMemorySchema
from skatepark.schemas.memory.summary_pipeline_memory_schema import SummaryPipelineMemorySchema

from skatepark.schemas.structures.structure_schema import StructureSchema
from skatepark.schemas.structures.pipeline_schema import PipelineSchema
from skatepark.schemas.structures.workflow_schema import WorkflowSchema

__all__ = [
    "BaseSchema",

    "PolymorphicSchema",

    "RuleSchema",

    "TiktokenTokenizerSchema",

    "PromptDriverSchema",
    "OpenAiPromptDriverSchema",

    "StepSchema",
    "PromptStepSchema",
    "ToolkitStepSchema",

    "SummarizerSchema",
    "PromptDriverSummarizerSchema",

    "PipelineRunSchema",
    "PipelineMemorySchema",
    "BufferPipelineMemorySchema",
    "SummaryPipelineMemorySchema",

    "StructureSchema",
    "PipelineSchema",
    "WorkflowSchema"
]



from skatepark.artifacts.structure_artifact import StructureArtifact
from skatepark.artifacts.error_output import ErrorOutput
from skatepark.artifacts.text_output import TextOutput


__all__ = [
    "StructureArtifact",
    "ErrorOutput",
    "TextOutput"
]



from skatepark.structures.structure import Structure
from skatepark.structures.pipeline import Pipeline
from skatepark.structures.workflow import Workflow


__all__ = [
    "Structure",
    "Pipeline",
    "Workflow"
]



from abc import abstractmethod
from marshmallow import fields
from skatepark.schemas import PolymorphicSchema, RuleSchema, BaseSchema


class StructureSchema(BaseSchema):
    class Meta:
        ordered = True

    id = fields.Str()
    type = fields.Str(required=True)
    prompt_driver = fields.Nested(PolymorphicSchema())
    rules = fields.List(fields.Nested(RuleSchema()))
    steps = fields.List(fields.Nested(PolymorphicSchema()))

    @abstractmethod
    def make_obj(self, data, **kwargs):
        ...



from marshmallow import fields, post_load
from skatepark.schemas import BaseSchema


class PipelineRunSchema(BaseSchema):
    class Meta:
        ordered = True

    id = fields.Str()
    input = fields.Str()
    output = fields.Str()

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.memory import PipelineRun

        return PipelineRun(**data)



import cohere
from attr import define, field
from skatepark.utils import Tokenizer


@define(frozen=True)
class CohereTokenizer(Tokenizer):
    DEFAULT_MODEL = "xlarge"
    MAX_TOKENS = 2048

    model: str = field(default=DEFAULT_MODEL, kw_only=True)
    stop_sequence: str = field(default=Tokenizer.DEFAULT_STOP_SEQUENCE, kw_only=True)
    client: cohere.Client = field(kw_only=True)

    @property
    def max_tokens(self) -> int:
        return self.MAX_TOKENS

    def tokens_left(self, text: str) -> int:
        diff = self.max_tokens - self.token_count(text)

        if diff > 0:
            return diff
        else:
            return 0

    def token_count(self, text: str) -> int:
        return len(self.encode(text))

    def encode(self, text: str) -> list[int]:
        return self.client.tokenize(text=text).tokens

    def decode(self, tokens: list[int]) -> str:
        return self.client.detokenize(tokens=tokens).text



from __future__ import annotations
import json
import logging
import uuid
from abc import ABC, abstractmethod
from logging import Logger
from typing import Optional, Union, TYPE_CHECKING
from attr import define, field, Factory
from rich.logging import RichHandler
from skatepark.drivers import PromptDriver, OpenAiPromptDriver
from skatepark.utils import J2, ToolLoader

if TYPE_CHECKING:
    from skatepark.rules import Rule
    from skatepark.steps import Step


@define
class Structure(ABC):
    LOGGER_NAME = "skatepark"

    id: str = field(default=Factory(lambda: uuid.uuid4().hex), kw_only=True)
    type: str = field(default=Factory(lambda self: self.__class__.__name__, takes_self=True), kw_only=True)
    prompt_driver: PromptDriver = field(default=OpenAiPromptDriver(), kw_only=True)
    rules: list[Rule] = field(factory=list, kw_only=True)
    steps: list[Step] = field(factory=list, kw_only=True)
    custom_logger: Optional[Logger] = field(default=None, kw_only=True)
    logger_level: int = field(default=logging.INFO, kw_only=True)
    tool_loader: ToolLoader = field(default=ToolLoader(), kw_only=True)

    _execution_args: tuple = ()
    _logger: Optional[Logger] = None

    def __attrs_post_init__(self):
        for step in self.steps:
            step.structure = self

    @property
    def execution_args(self) -> tuple:
        return self._execution_args

    @property
    def logger(self) -> Logger:
        if self.custom_logger:
            return self.custom_logger
        else:
            if self._logger is None:
                self._logger = logging.getLogger(self.LOGGER_NAME)

                self._logger.propagate = False
                self._logger.level = self.logger_level

                self._logger.handlers = [
                    RichHandler(
                        show_time=True,
                        show_path=False
                    )
                ]

            return self._logger

    def is_finished(self) -> bool:
        return all(s.is_finished() for s in self.steps)

    def is_executing(self) -> bool:
        return any(s for s in self.steps if s.is_executing())

    def is_empty(self) -> bool:
        return not self.steps

    def find_step(self, step_id: str) -> Optional[Step]:
        return next((step for step in self.steps if step.id == step_id), None)

    def add_steps(self, *steps: Step) -> list[Step]:
        return [self.add_step(s) for s in steps]

    def prompt_stack(self, step: Step) -> list[str]:
        from skatepark.steps import ToolkitStep

        tools = step.tools if isinstance(step, ToolkitStep) else []

        stack = [
            J2("prompts/context.j2").render(
                rules=self.rules,
                tool_names=str.join(", ", [tool.name for tool in tools]),
                tools=[J2("prompts/tool.j2").render(tool=tool) for tool in tools]
            )
        ]

        return stack

    def to_prompt_string(self, step: Step) -> str:
        return self.stack_to_prompt_string(self.prompt_stack(step))

    def stack_to_prompt_string(self, stack: list[str]) -> str:
        return str.join("\n", stack)

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), indent=2)

    def context(self, step: Step) -> dict[str, any]:
        return {
            "args": self.execution_args,
            "structure": self,
        }

    @abstractmethod
    def add_step(self, step: Step) -> Step:
        ...

    @abstractmethod
    def run(self, *args) -> Union[Step, list[Step]]:
        ...

    @abstractmethod
    def to_dict(self) -> dict:
        ...

    @classmethod
    @abstractmethod
    def from_dict(cls, workflow_dict: dict) -> Structure:
        ...

    @classmethod
    @abstractmethod
    def from_json(cls, workflow_json: str) -> Structure:
        ...



from skatepark.steps.step import Step
from skatepark.steps.prompt_step import PromptStep
from skatepark.steps.tool_substep import ToolSubstep
from skatepark.steps.toolkit_step import ToolkitStep

__all__ = [
    "Step",
    "PromptStep",
    "ToolSubstep",
    "ToolkitStep"
]



from __future__ import annotations
import logging
import time
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from attr import define, field, Factory
from skatepark.utils import Tokenizer

if TYPE_CHECKING:
    from skatepark.artifacts import TextOutput


@define
class PromptDriver(ABC):
    max_retries: int = field(default=8, kw_only=True)
    retry_delay: float = field(default=1, kw_only=True)
    type: str = field(default=Factory(lambda self: self.__class__.__name__, takes_self=True), kw_only=True)
    temperature: float = field(default=0.5, kw_only=True)
    model: str
    tokenizer: Tokenizer

    def run(self, **kwargs) -> TextOutput:
        for attempt in range(0, self.max_retries + 1):
            try:
                return self.try_run(**kwargs)
            except Exception as e:
                logging.error(f"PromptDriver.run attempt {attempt} failed: {e}\nRetrying in {self.retry_delay} seconds")

                if attempt < self.max_retries:
                    time.sleep(self.retry_delay)
                else:
                    raise e

    @abstractmethod
    def try_run(self, **kwargs) -> TextOutput:
        ...



from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Optional
from attr import define, field
from skatepark.artifacts import StructureArtifact


if TYPE_CHECKING:
    from skatepark.steps import Step


@define(frozen=True)
class ErrorOutput(StructureArtifact):
    exception: Optional[Exception] = field(default=None, kw_only=True)
    step: Optional[Step] = field(default=None, kw_only=True)



from __future__ import annotations
from typing import TYPE_CHECKING
from attr import define, field

if TYPE_CHECKING:
    from skatepark.memory import PipelineMemory


@define(frozen=True)
class Conversation:
    memory: PipelineMemory = field()

    def lines(self) -> list[str]:
        lines = []

        for run in self.memory.runs:
            lines.append(f"Q: {run.input}")
            lines.append(f"A: {run.output}")

        return lines

    def to_string(self) -> str:
        return str.join("\n", self.lines())



import cohere
from attr import define, field, Factory
from skatepark.artifacts import TextOutput
from skatepark.drivers import PromptDriver
from skatepark.utils import CohereTokenizer


@define
class CoherePromptDriver(PromptDriver):
    api_key: str = field(kw_only=True)
    model: str = field(default=CohereTokenizer.DEFAULT_MODEL, kw_only=True)
    client: cohere.Client = field(
        default=Factory(lambda self: cohere.Client(self.api_key), takes_self=True), kw_only=True
    )
    tokenizer: CohereTokenizer = field(
        default=Factory(lambda self: CohereTokenizer(model=self.model, client=self.client), takes_self=True),
        kw_only=True
    )

    def try_run(self, value: any) -> TextOutput:
        result = self.client.generate(
            value,
            model=self.model,
            temperature=self.temperature,
            end_sequences=[self.tokenizer.stop_sequence, "Input:"],
            max_tokens=self.tokenizer.tokens_left(value)
        )

        if len(result.generations) == 1:
            generation = result.generations[0]

            return TextOutput(
                value=generation.text,
                meta=result.meta
            )
        else:
            raise Exception("Completion with more than one choice is not supported yet.")


from __future__ import annotations
import json
from typing import TYPE_CHECKING
from typing import Optional
from attr import define, field
from skatepark.utils import J2
from skatepark.memory import PipelineMemory

if TYPE_CHECKING:
    from skatepark.summarizers import Summarizer
    from skatepark.memory import PipelineRun


@define
class SummaryPipelineMemory(PipelineMemory):
    offset: int = field(default=1, kw_only=True)
    summarizer: Optional[Summarizer] = field(default=None, kw_only=True)
    summary: Optional[str] = field(default=None, kw_only=True)
    summary_index: int = field(default=0, kw_only=True)

    def unsummarized_runs(self, last_n: Optional[int] = None) -> list[PipelineRun]:
        summary_index_runs = self.runs[self.summary_index:]

        if last_n:
            last_n_runs = self.runs[-last_n:]

            if len(summary_index_runs) > len(last_n_runs):
                return last_n_runs
            else:
                return summary_index_runs
        else:
            return summary_index_runs

    def process_add_run(self, run: PipelineRun) -> None:
        super().process_add_run(run)

        if self.summarizer:
            unsummarized_runs = self.unsummarized_runs()
            runs_to_summarize = unsummarized_runs[:max(0, len(unsummarized_runs) - self.offset)]

            if len(runs_to_summarize) > 0:
                self.summary = self.summarizer.summarize(self, runs_to_summarize)
                self.summary_index = 1 + self.runs.index(runs_to_summarize[-1])

    def to_prompt_string(self, last_n: Optional[int] = None):
        return J2("prompts/memory.j2").render(
            summary=self.summary,
            runs=self.unsummarized_runs(last_n)
        )

    def to_dict(self) -> dict:
        return SummaryPipelineMemory().dump(self)

    @classmethod
    def from_dict(cls, memory_dict: dict) -> PipelineMemory:
        return SummaryPipelineMemory().load(memory_dict)

    @classmethod
    def from_json(cls, memory_json: str) -> PipelineMemory:
        return SummaryPipelineMemory.from_dict(json.loads(memory_json))



from skatepark.rules.rule import Rule


def be_truthful() -> Rule:
    return Rule(
        "be truthful and say \"I don't know\" if you don't have the knowledge to answer a question"
    )


def speculate() -> Rule:
    return Rule(
        "say \"I don't know\" if you don't know the answer to the question but also be creative and speculate what the "
        "possible answer could be"
    )


def your_name_is(name: str) -> Rule:
    return Rule(
        f"respond to name \"{name}\""
    )



from __future__ import annotations
import json
from attr import define, field
from skatepark.memory import PipelineMemory, PipelineRun


@define
class BufferPipelineMemory(PipelineMemory):
    buffer_size: int = field(default=1, kw_only=True)

    def process_add_run(self, run: PipelineRun) -> None:
        super().process_add_run(run)

        while len(self.runs) > self.buffer_size:
            self.runs.pop(0)

    def to_dict(self) -> dict:
        return BufferPipelineMemory().dump(self)

    @classmethod
    def from_dict(cls, memory_dict: dict) -> PipelineMemory:
        return BufferPipelineMemory().load(memory_dict)

    @classmethod
    def from_json(cls, memory_json: str) -> PipelineMemory:
        return BufferPipelineMemory.from_dict(json.loads(memory_json))



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional
from attr import define, field
from griptape.core import BaseTool
from skatepark.utils import J2
from skatepark.steps import PromptStep
from skatepark.artifacts import TextOutput, ErrorOutput

if TYPE_CHECKING:
    from skatepark.steps import ToolSubstep


@define
class ToolkitStep(PromptStep, ABC):
    DEFAULT_MAX_SUBSTEPS = 20

    tool_names: list[str] = field(kw_only=True)
    max_substeps: int = field(default=DEFAULT_MAX_SUBSTEPS, kw_only=True)
    _substeps: list[ToolSubstep] = field(factory=list)

    @property
    def tools(self) -> list[BaseTool]:
        return [
            t for t in [self.structure.tool_loader.load_tool(t) for t in self.tool_names] if t is not None
        ]

    def run(self) -> TextOutput:
        from skatepark.steps import ToolSubstep

        self._substeps.clear()

        substep = self.add_substep(
            ToolSubstep(
                self.active_driver().run(value=self.structure.to_prompt_string(self)).value
            )
        )

        while True:
            if substep.output is None:
                if len(self._substeps) >= self.max_substeps:
                    substep.output = ErrorOutput(
                        f"Exceeded maximum tool execution limit of {self.max_substeps} per step",
                        step=self
                    )
                elif substep.tool_name is None:
                    # handle case when the LLM failed to follow the ReAct prompt and didn't return a proper action
                    substep.output = TextOutput(substep.prompt_template)
                else:
                    substep.before_run()
                    substep.run()
                    substep.after_run()

                    substep = self.add_substep(
                        ToolSubstep(
                            self.active_driver().run(value=self.structure.to_prompt_string(self)).value
                        )
                    )
            else:
                break

        self.output = substep.output

        return self.output

    def render(self) -> str:
        return J2("prompts/steps/tool/tool.j2").render(
            step=self,
            substeps=self._substeps
        )

    def find_substep(self, step_id: str) -> Optional[ToolSubstep]:
        return next((step for step in self._substeps if step.id == step_id), None)

    def add_substep(self, substep: ToolSubstep) -> ToolSubstep:
        substep.attach(self)

        if len(self._substeps) > 0:
            self._substeps[-1].add_child(substep)

        self._substeps.append(substep)

        return substep

    def find_tool(self, tool_name: str) -> Optional[BaseTool]:
        return next(
            (t for t in self.tools if t.name == tool_name),
            None
        )



from skatepark.rules.rule import Rule


def return_valid_json() -> Rule:
    return Rule(
        "only output valid JSON"
    )


def return_array() -> Rule:
    return Rule(
        "only output a valid JSON array"
    )


def return_object() -> Rule:
    return Rule(
        "only output a valid JSON object"
    )


def put_answer_in_field(value: str) -> Rule:
    return Rule(
        f"only output a valid JSON object with your answer in '{value}'"
    )



from marshmallow import fields, post_load
from skatepark.schemas import BaseSchema


class TiktokenTokenizerSchema(BaseSchema):
    model = fields.Str()
    stop_sequence = fields.Str()

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.utils import TiktokenTokenizer

        return TiktokenTokenizer(**data)



from marshmallow import fields, post_load
from skatepark.schemas import StepSchema, PolymorphicSchema


class PromptStepSchema(StepSchema):
    prompt_template = fields.Str(required=True)
    context = fields.Dict(keys=fields.Str(), values=fields.Raw())
    driver = fields.Nested(PolymorphicSchema(), allow_none=True)

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.steps import PromptStep

        return PromptStep(**data)



from marshmallow import fields, post_load
from skatepark.schemas import PromptDriverSchema


class OpenAiPromptDriverSchema(PromptDriverSchema):
    api_type = fields.Str()
    api_version = fields.Str(allow_none=True)
    api_base = fields.Str()
    api_key = fields.Str(allow_none=True)
    organization = fields.Str(allow_none=True)
    model = fields.Str()
    temperature = fields.Float()
    user = fields.Str()

    @post_load
    def make_obj(self, data, **kwargs):
        from skatepark.drivers import OpenAiPromptDriver

        return OpenAiPromptDriver(**data)



from typing import Optional
from attr import define, field
from skatepark.artifacts import StructureArtifact
from skatepark.utils import TiktokenTokenizer, Tokenizer


@define(frozen=True)
class TextOutput(StructureArtifact):
    meta: Optional[any] = field(default=None)
    tokenizer: Tokenizer = field(default=TiktokenTokenizer(), kw_only=True)

    def token_count(self) -> Optional[int]:
        if isinstance(self.value, str):
            return self.tokenizer.token_count(self.value)
        else:
            return None



from __future__ import annotations
import json
from typing import TYPE_CHECKING, Optional
from attr import define, field
from skatepark.artifacts import ErrorOutput
from skatepark.structures import Structure
from skatepark.memory import PipelineMemory, PipelineRun
from skatepark.utils import J2

if TYPE_CHECKING:
    from skatepark.steps import Step


@define
class Pipeline(Structure):
    memory: Optional[PipelineMemory] = field(default=None, kw_only=True)
    autoprune_memory: bool = field(default=True, kw_only=True)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        if self.memory:
            self.memory.pipeline = self

    def first_step(self) -> Optional[Step]:
        return None if self.is_empty() else self.steps[0]

    def last_step(self) -> Optional[Step]:
        return None if self.is_empty() else self.steps[-1]

    def finished_steps(self) -> list[Step]:
        return [s for s in self.steps if s.is_finished()]

    def add_step(self, step: Step) -> Step:
        if self.last_step():
            self.last_step().add_child(step)
        else:
            step.structure = self

            self.steps.append(step)

        return step

    def prompt_stack(self, step: Step) -> list[str]:
        final_stack = super().prompt_stack(step)
        step_prompt = J2("prompts/pipeline.j2").render(
            has_memory=self.memory is not None,
            finished_steps=self.finished_steps(),
            current_step=step
        )

        if self.memory:
            if self.autoprune_memory:
                last_n = len(self.memory.runs)
                should_prune = True

                while should_prune and last_n > 0:
                    temp_stack = final_stack.copy()
                    temp_stack.append(step_prompt)

                    temp_stack.append(self.memory.to_prompt_string(last_n))

                    if self.prompt_driver.tokenizer.tokens_left(self.stack_to_prompt_string(temp_stack)) > 0:
                        should_prune = False
                    else:
                        last_n -= 1

                if last_n > 0:
                    final_stack.append(self.memory.to_prompt_string(last_n))
            else:
                final_stack.append(self.memory.to_prompt_string())

        final_stack.append(step_prompt)

        return final_stack

    def run(self, *args) -> Step:
        self._execution_args = args

        [step.reset() for step in self.steps]

        self.__run_from_step(self.first_step())

        if self.memory:
            run = PipelineRun(
                input=self.first_step().render_prompt(),
                output=self.last_step().output.value
            )

            self.memory.add_run(run)

        self._execution_args = ()

        return self.last_step()

    def context(self, step: Step) -> dict[str, any]:
        context = super().context(step)

        context.update(
            {
                "input": step.parents[0].output.value if step.parents and step.parents[0].output else None,
                "parent": step.parents[0] if step.parents else None,
                "child": step.children[0] if step.children else None
            }
        )

        return context

    def to_dict(self) -> dict:
        from skatepark.schemas import PipelineSchema

        return PipelineSchema().dump(self)

    @classmethod
    def from_dict(cls, pipeline_dict: dict) -> Pipeline:
        from skatepark.schemas import PipelineSchema

        return PipelineSchema().load(pipeline_dict)

    @classmethod
    def from_json(cls, pipeline_json: str) -> Pipeline:
        return Pipeline.from_dict(json.loads(pipeline_json))

    def __run_from_step(self, step: Optional[Step]) -> None:
        if step is None:
            return
        else:
            if isinstance(step.execute(), ErrorOutput):
                return
            else:
                self.__run_from_step(next(iter(step.children), None))
